/*
 * actF.cpp
 * Contains definitions for checking action Fs.
 *
 * Copyright 2005-2006 by Dale McCoy.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include<string>
#include<cassert>

using namespace std;

#include"renum.h"
#include"inlines.h"
#include"pseudo.h"
#include"messages.h"
#include"ExpandingArray.h"
#include"strings.h"

class IDarray{
public:
    void init(){used.resize(0);sprite.resize(0);}
    bool is_defined(int id)const{return sprite[id]!=0;}
    bool is_used(int id)const{return used[id];}
    unsigned short defined_at(int id)const{return sprite[id];}
    void define(unsigned int id){
        used[id]=false;
        sprite[id]=(unsigned short)_spritenum;
    }
    void use(int id){used[id]=true;}
protected:
    Expanding0Array<unsigned short>sprite;
    Expanding0Array<bool>used;
}status;
static const IDarray&crStatus=status;

void InitF(){status.init();}

void CheckF(PseudoSprite&data){
    data.SetAllHex();
    uint id=data.ExtractByte(1),offset=2,langs,oldoff=0;
    uint offs[5]={0,0,0,0,0};
    if(status.is_defined(id)&&!status.is_used(id))IssueMessage(WARNING1,UNUSED_ID,id&0x7F,status.defined_at(id));
    status.define(id&0x7F);
    if(id&0x80)status.use(id&0x7F);
    bool defall=false;
    if(id&0x80){
        uint names=0;
        while((langs=data.ExtractByte(offset))!=0){
            names++;
            if(names>1)data.SetEol(oldoff-1);
            oldoff=offset;
            if(!(langs&0x40)){
                if(langs&0x10){if(offs[4])IssueMessage(WARNING2,DUPLICATE_LANG_NAME,offset,4,offs[4]);offs[4]=offset;}
                if(langs&0x08){if(offs[3])IssueMessage(WARNING2,DUPLICATE_LANG_NAME,offset,3,offs[3]);offs[3]=offset;}
                if(langs&0x04){if(offs[2])IssueMessage(WARNING2,DUPLICATE_LANG_NAME,offset,2,offs[2]);offs[2]=offset;}
                if(langs&0x02){if(offs[1])IssueMessage(WARNING2,DUPLICATE_LANG_NAME,offset,1,offs[1]);offs[1]=offset;}
                if(langs&0x01){if(offs[0])IssueMessage(WARNING2,DUPLICATE_LANG_NAME,offset,0,offs[0]);offs[0]=offset;}
            }else if((langs&0x3F)==0x3F)defall=true;
            else if((langs&0x3F)<5){
                if(offs[langs&0x3F])IssueMessage(WARNING2,DUPLICATE_LANG_NAME,offset,langs&0x3F,offs[langs&0x3F]);
                offs[langs&0x3F]=offset;
            }
            if(CheckString(data,++offset,0)){
                IssueMessage(FATAL,OVERRAN_F_NAME,oldoff+1,langs);
                return;
            }
        }
        if(names>1)data.SetEol(oldoff-1);
        if(!defall&&!offs[4])IssueMessage(WARNING1,MISSING_LANG_NAME,4);
        if(!defall&&!offs[3])IssueMessage(WARNING1,MISSING_LANG_NAME,3);
        if(!defall&&!offs[2])IssueMessage(WARNING1,MISSING_LANG_NAME,2);
        if(!defall&&!offs[1])IssueMessage(WARNING1,MISSING_LANG_NAME,1);
        if(!defall&&!offs[0])IssueMessage(WARNING1,MISSING_LANG_NAME,0);
        offset++;
    }
    uint num_parts=data.SetEol(offset).ExtractByte(offset);
    if(id&0x80)data.SetEol(offset-1);
    if(!num_parts)IssueMessage(ERROR,NO_PARTS,offset);
    for(uint i=0;i<num_parts;i++){
        uint textcount=data.ExtractByte(++offset);
        if(!textcount)IssueMessage(ERROR,NO_PARTS,offset);
        uint total_prob=0,firstbit=data.ExtractByte(++offset),fb_offs=offset,numbits=data.ExtractByte(++offset);
        data.SetEol(offset,true);
        if(firstbit+numbits>32)IssueMessage(ERROR,OUT_OF_RANGE_BITS,32,fb_offs);
        for(uint j=0;j<textcount;j++){
            uint prob=data.ExtractByte(++offset);
            total_prob+=prob&0x7F;
            if(!(prob&0x7F))IssueMessage(WARNING1,NO_PROBABILITY,offset);
            if(prob&0x80){
                uint id=data.ExtractByte(++offset)&0x7F;
                if(!status.is_defined(id))IssueMessage(ERROR,UNDEFINED_ID,offset,id);
                else status.use(id);
                data.SetEol(offset,true);
            }else{
                oldoff=++offset;
                if(CheckString(data,offset,0)){
                    IssueMessage(FATAL,OVERRAN_F_PART,oldoff);
                    return;
                }
                data.SetEol(--offset,true);
            }
        }
        uint minbits=1;
        while(total_prob>>=1)minbits++;
        if(numbits<minbits)IssueMessage(WARNING1,INSUFFICIENT_BITS,fb_offs+1,minbits,numbits);
    }
    if(++offset!=data.Length())IssueMessage(WARNING2,EXTRA_DATA,offset);
}

void finalF(){
    bool header=false;
    for(uint i=0;i<128;i++)
        if(crStatus.is_defined(i)&&!crStatus.is_used(i)){
            if(!header){
                IssueMessage(WARNING1,UNUSEDFIDLEAD,i);
                IssueMessage(WARNING1,UNEXP_EOF_TOWNNAMES,i);
                header=true;
            }
            IssueMessage(WARNING1,UNUSEDIDFINAL,i,crStatus.defined_at(i));
        }
}
